#include "notify_manage.h"
#include "notify_manage_communication.h"
#include "string_convert.h"
#include <jsoncpp/json/json.h>
#include <sys/stat.h>

Notify_manager::Notify_manager()
{

}
Notify_manager::~Notify_manager()
{
	notify_manager_uninit();
}

//初始化
Error_manager Notify_manager::notify_manager_init()
{
	LOG(INFO) << " ---Notify_manager::communication_init() run--- "<< this;
	m_notify_manager_satus = eNotify_manager_unknown;
	//LED屏初始化
	m_led_protocol.led_protocol_init();
	//设置LED屏矩形 分成三个区域
	m_led_protocol.set_show_rectangle(0,0,0,133,60);
	m_led_protocol.set_show_rectangle(1,133,0,133,60);
	m_led_protocol.set_show_rectangle(2,267,0,133,60);

//	m_led_protocol.set_show_rectangle(0,0,0,50,19);
//	m_led_protocol.set_show_rectangle(1,55,0,170,19);
//	m_led_protocol.set_show_rectangle(2,226,0,170,19);
	//设置LED屏移动模式 0x05是为向上移动
	m_led_protocol.set_display_mode(0,0x05);
	m_led_protocol.set_display_mode(1,0x05);
	m_led_protocol.set_display_mode(2,0x05);

	//从配置文件读取IP
	if(!  proto_tool::read_proto_param(SINGLE_CHIP_COMMUNICATION_PARAMETER_PATH,m_notify_communication_parameter_all) )
	{
		return Error_manager(COMMUNICATION_READ_PROTOBUF_ERROR,MINOR_ERROR,
							 "Notify_manager::communication_socket_base read_proto_param  failed");
	}
	return notify_manager_init_from_protobuf(m_notify_communication_parameter_all);
}
Error_manager Notify_manager::notify_manager_init_from_protobuf(Communication_proto::Notify_communication_parameter_all& t_notify_communication_parameter_all)
{
	LOG(INFO) << " ---Notify_manager::communication_init_from_protobuf() run--- "<< this;
	Error_manager t_error;
	//判断配置文件IP数量和端口数量是否一致
	if(t_notify_communication_parameter_all.notify_communication_parameters().singlechip_ip_vector_size() !=
	   t_notify_communication_parameter_all.notify_communication_parameters().singlechip_port_vector_size())
	{
		return Error_manager(COMMUNICATION_READ_PROTOBUF_ERROR,MINOR_ERROR,
							 "singlechip_ip_vector_size() ！= singlechip_port_vector_size()   failed");
	}
	//有多少个IP添加多少个通信
	for(int i=0;i<t_notify_communication_parameter_all.notify_communication_parameters().singlechip_ip_vector_size();++i)
	{
		//填充单片机通信
		m_singlechip_communication_vector.push_back(Communication_tcp());
		m_singlechip_communication_vector[i].communication_tcp_init();
		m_singlechip_data_vector.push_back(message::Singlechip_data());

		std::string t_ip = t_notify_communication_parameter_all.notify_communication_parameters().singlechip_ip_vector(i).c_str();
		unsigned short t_port = t_notify_communication_parameter_all.notify_communication_parameters().singlechip_port_vector(i);

		//连接IP
		bool result = m_singlechip_communication_vector[i].communication_tcp_connect( t_ip.c_str(),t_port);
		if (  result )
		{
			LOG(INFO) << "-------- "<< i << " 号单片机连接成功! -------- ip = "<<t_ip<<" -------- port = "<<t_port<< " --------" <<this;
		}
		else
		{
			LOG(INFO) << "######## "<< i << " 号单片机连接失败! ######## ip = "<<t_ip<<" ######## port = "<<t_port<< " ########" <<this;

		}
	}

	for (int i = 0; i < UNIT_COUNT; ++i)
	{
		m_dispatch_terminal_status_msg_vector.push_back(message::Dispatch_terminal_status_msg());
		m_dispatch_pickup_process_updata_time_vector.push_back("");
	}
	//led屏初始化
	send_led_showstring(0,"");
	send_led_showstring(1,"");
	send_led_showstring(2,"");
	//启动
	//test();
	notify_manager_run();
	return Error_code::SUCCESS;
}
Error_manager Notify_manager::notify_manager_run()
{

	// 初始化任务处理线程池
	m_thread_pool.thread_pool_init(10);
	m_notify_manager_satus = eNotify_manager_normal;

	//接受线程默认循环, 内部的nn_recv进行等待, 超时1ms 每一个通信都有一个线程
	for (int i = 0; i < m_singlechip_communication_vector.size(); ++i)
	{
		m_receive_condition_vector.push_back(new Thread_condition());
		mp_receive_singlechip_data_thread_vector.push_back(new std::thread());
		m_receive_condition_vector[i]->reset(false, false, false);
		mp_receive_singlechip_data_thread_vector[i] = new std::thread(&Notify_manager::receive_singlechip_data_thread, this,i);
	}

	//启动自动重连线程
	m_reconnect_condition.reset(false, false, false);
	mp_reconnect_thread= new std::thread(&Notify_manager::reconnect_thread, this);

	return Error_code::SUCCESS;
}
void Notify_manager::test()
{

	send_led_showstring(0,"\\C7鄂WK0001 \\C1排队2");
	send_led_showstring(1,"\\C7鄂WK0002 \\C3工作中");
	send_led_showstring(2,"\\C7鄂WK0003 \\C25号口");


}
//自动重连线程
void Notify_manager::reconnect_thread()
{
	LOG(INFO) << " Notify_manager::reconnect_thread  start "<< this;

	//客户端连接会自动重连 因此这里只需要检查服务端是否断开
	while (m_reconnect_condition.is_alive())
	{
		m_reconnect_condition.wait_for_ex(std::chrono::milliseconds(3000));
		if (m_reconnect_condition.is_alive())
		{
			std::this_thread::yield();
			for (int i = 0; i < m_singlechip_communication_vector.size(); ++i)
			{
				//检查服务端连接是否正常
				if ( !m_singlechip_communication_vector[i].is_connected() )
				{
					//关闭此前的连接
					m_singlechip_communication_vector[i].communication_tcp_uninit();
					//重新初始化连接
					m_singlechip_communication_vector[i].communication_tcp_init();
					std::string t_ip = m_notify_communication_parameter_all.notify_communication_parameters().singlechip_ip_vector(i).c_str();
					unsigned short t_port = m_notify_communication_parameter_all.notify_communication_parameters().singlechip_port_vector(i);
					//尝试重连
					bool result = m_singlechip_communication_vector[i].communication_tcp_connect(t_ip.c_str(),t_port);
					if (  result )
					{
						LOG(INFO) << "-------- "<< i << " 号单片机重连成功! -------- ip = "<<t_ip<<" -------- port = "<<t_port<< " --------" <<this;
					}
					else
					{
						LOG(WARNING) << "######## "<< i << " 号单片机重连失败! ######## ip = "<<t_ip<<" ######## port = "<<t_port<< " ########" <<this;
					}
				}

			}

		}
	}
	LOG(INFO) << " Notify_manager::reconnect_thread  end "<< this;
}
//接收单片机数据线程
void Notify_manager::receive_singlechip_data_thread(int singlechip_id)
{
	LOG(INFO) << " Notify_manager::receive_singlechip_data_thread "<<singlechip_id<<" start "<< this;

	//设置为阻塞模式
	m_singlechip_communication_vector[singlechip_id].set_block(false);
	//通信接受线程, 负责接收消息
	while (m_receive_condition_vector[singlechip_id]->is_alive())
	{
		m_receive_condition_vector[singlechip_id]->wait_for_ex(std::chrono::milliseconds(1));
		if ( m_receive_condition_vector[singlechip_id]->is_alive() )
		{
			std::this_thread::yield();
			char buf[1024];
			memset(buf, 0, 1024);
			//循环接收消息
			int length = m_singlechip_communication_vector[singlechip_id].communication_tcp_recv(buf, sizeof(buf));
			//大于有效数据
			if( length > 0)
			{
				//处理数据
				std::string t_data = data_processing(buf,singlechip_id);
				//说明为解析的完整数据包
				if ( t_data != "" )
				{
					//std::cout << " data = " << t_data <<" ------singlechip_id = " <<singlechip_id<< std::endl;
					//json转protobuf
					message::Singlechip_data singlechipData;
					google::protobuf::util::JsonStringToMessage(t_data,&singlechipData);

					//发送出去
					message::Notify_status_msg notifyStatusMsg;
					message::Base_info baseInfo;
					baseInfo.set_sender(message::Communicator::eNotify);
					baseInfo.set_receiver(message::Communicator::eEmpty);
					baseInfo.set_msg_type(message::Message_type::eNotify_status_msg);
					baseInfo.set_timeout_ms(1000);

					message::Id_struct idStruct;
					idStruct.set_terminal_id(singlechip_id);

					message::Error_manager errorManager;
					errorManager.set_error_code(0);

					notifyStatusMsg.mutable_base_info()->CopyFrom(baseInfo);
					notifyStatusMsg.mutable_id_struct()->CopyFrom(idStruct);
					notifyStatusMsg.mutable_error_manager()->CopyFrom(errorManager);
					notifyStatusMsg.mutable_singlechipdata()->CopyFrom(singlechipData);
					notifyStatusMsg.set_singlechip_validity(true);

					Notify_manage_communication::get_instance_references().encapsulate_msg(notifyStatusMsg.SerializeAsString());

					//std::cout << " singlechipData.DebugString() = " << singlechipData.DebugString() << std::endl;
					//保存数据
					{//上锁
						std::unique_lock<std::mutex> lk(m_singlechip_data_mutex);
						m_singlechip_data_vector[singlechip_id] = singlechipData;
					}
				}
			}

		}
	}

	LOG(INFO) << " Notify_manager::receive_singlechip_data_thread "<<singlechip_id<<" end "<< this;
	return;
}
//处理粘包
std::string Notify_manager::data_processing(char *data,int singlechip_id)
{
	std::string t_data = data;
	//从传入数据查找包头
	char* head = (char*)memchr(t_data.c_str(),'@',t_data.length());
	if ( head != NULL  )
	{
		//找到包头查找包尾
		std::string t_head  = head;
		//查找包尾
		int tail_index = t_head.find('$');
		if ( tail_index != -1 )
		{
			//找到包尾 丢弃包尾后面的数据 的到包头+完整数据
			std::string baotou_data = t_head.substr(0,tail_index);
			//读取完整数据的长度
			int length = std::stoi(baotou_data.substr(1,DATA_HEADER_LENGTH-1));
			//读取完整数据
			std::string complete_data = baotou_data.substr(DATA_HEADER_LENGTH,length).c_str();
			//若读出来的完整数据长度与解析出来的数据长度一致则说明数据正确
			if ( complete_data.length() == length)
			{
				//返回完整数据
				return complete_data;
			}
		}
	}
	return "";
}
//反初始化
Error_manager Notify_manager::notify_manager_uninit()
{
	//关闭线程池
	m_thread_pool.thread_pool_uninit();
	m_notify_manager_satus = eNotify_manager_unknown;
	//回收线程资源
	if (mp_reconnect_thread)
	{
		m_reconnect_condition.kill_all();
	}
	if (mp_reconnect_thread)
	{
		mp_reconnect_thread->join();
		delete mp_reconnect_thread;
		mp_reconnect_thread = NULL;
	}
	//关闭通信 回收线程资源
	for (int i = 0; i < m_singlechip_communication_vector.size(); ++i)
	{
		m_singlechip_communication_vector[i].communication_tcp_uninit();
		if (mp_receive_singlechip_data_thread_vector[i])
		{
			m_receive_condition_vector[i]->kill_all();
		}

		if (mp_receive_singlechip_data_thread_vector[i])
		{
			mp_receive_singlechip_data_thread_vector[i]->join();
			delete mp_receive_singlechip_data_thread_vector[i];
			mp_receive_singlechip_data_thread_vector[i] = NULL;
		}
	}
	return Error_code::SUCCESS;
}

Error_manager Notify_manager::check_status()
{
	if(m_notify_manager_satus != eNotify_manager_normal)
	{
		return Error_manager(Error_code::ERROR, Error_level::MAJOR_ERROR,
							 " Notify_manager::check_executer error ");
	}
	return SUCCESS;
}
//处理消息的执行函数
Error_manager Notify_manager::execute_msg(Communication_message* p_msg)
{
	if ( p_msg == nullptr )
	{
		return Error_manager(Error_code::POINTER_IS_NULL, Error_level::MINOR_ERROR,
							 "  POINTER IS NULL ");
	}
	switch ( p_msg->get_message_type() )
	{
		//调度模块终端出入口的状态消息
		case Communication_message::eDispatch_status_msg:
		{
			message::Dispatch_terminal_status_msg msg;
			bool result = msg.ParseFromString(p_msg->get_message_buf());
			if(!result)
			{
				return Error_manager(Error_code::PARKSPACE_ALLOCATOR_MSG_PARSE_ERROR, Error_level::MINOR_ERROR,
									 " message::Dispatch_terminal_status_msg  ParseFromString error ");
			}
			m_thread_pool.enqueue(&Notify_manager::execute_for_dispatch_terminal_status_msg, this, msg);
			return SUCCESS;
		}
		//取车等候请求
		case Communication_message::eNotify_request_msg:
		{
			message::Notify_request_msg request;
			bool result = request.ParseFromString(p_msg->get_message_buf());
			if(!result)
			{
				return Error_manager(Error_code::PARKSPACE_ALLOCATOR_MSG_PARSE_ERROR, Error_level::MINOR_ERROR,
									 " message::eNotify_request_msg  ParseFromString error ");
			}
			m_thread_pool.enqueue(&Notify_manager::execute_for_notify_request, this, request);
			return SUCCESS;
		}
	}
	return Error_manager(Error_code::ERROR, Error_level::MINOR_ERROR,
						 " fun error ");
}
//调度模块终端出入口的状态消息
void Notify_manager::execute_for_dispatch_terminal_status_msg(message::Dispatch_terminal_status_msg msg)
{
	int t_unit_id = msg.id_struct().unit_id();
	//更新时间有变化则更新LED屏
	if ( m_dispatch_pickup_process_updata_time_vector[t_unit_id] != msg.dispatch_pickup_process_information_updata_time() )
	{
		//std::cout << "---------更新时间发生变化-----------------msg.dispatch_pickup_process_information_updata_time() = " << msg.dispatch_pickup_process_information_updata_time() << std::endl;
		m_dispatch_pickup_process_updata_time_vector[t_unit_id] = msg.dispatch_pickup_process_information_updata_time();
		m_dispatch_terminal_status_msg_vector[t_unit_id].CopyFrom(msg) ;
		//更新LED屏幕
		std::string led_show_string = "";

		for (int i = 0; i < m_dispatch_terminal_status_msg_vector[t_unit_id].dispatch_pickup_process_information_vector().size(); ++i)
		{
			//车牌号
			std::string t_car_numberplate = "\\C7"+m_dispatch_terminal_status_msg_vector[t_unit_id].dispatch_pickup_process_information_vector(i).car_numberplate();
			//状态
			std::string t_status_string = "";
			switch (  m_dispatch_terminal_status_msg_vector[t_unit_id].dispatch_pickup_process_information_vector(i).dispatch_process_status() )
			{
				//已完成
				case message::Dispatch_process_status::E_DISPATCH_POCESS_OVER:
				{
					t_status_string = "\\C2"+std::to_string(m_dispatch_terminal_status_msg_vector[t_unit_id].dispatch_pickup_process_information_vector(i).id_struct().terminal_id()+1)+"号口";
					break;
				}
				//工作中
				case message::Dispatch_process_status::E_DISPATCH_POCESS_WORKING:
				{
					t_status_string = "\\C3搬运中";
					break;
				}
				//等待中
				case message::Dispatch_process_status::E_DISPATCH_POCESS_WAIT:
				{
					t_status_string = "\\C1排队"+std::to_string(m_dispatch_terminal_status_msg_vector[t_unit_id].dispatch_pickup_process_information_vector(i).m_wait_number());
					break;
				}
				default:
					break;
			}
			led_show_string = led_show_string + t_car_numberplate + " " + t_status_string+"\\n";
		}
		//更新LED屏
		send_led_showstring(t_unit_id,led_show_string);
		LOG(INFO) << "更新时间有变化	更新LED屏	led_show_string = " <<led_show_string ;

		//std::cout<< " t_unit_id = " << t_unit_id  << " led_show_string = " << led_show_string << std::endl;
	}
}
//执行取车等候请求消息
void Notify_manager::execute_for_notify_request(message::Notify_request_msg request)
{
	//接到此消息就开门
	message::terminal_msg terminalMsg;
	terminalMsg.set_terminalid(request.id_struct().terminal_id());
	terminalMsg.set_dispatchdirection(2);
	terminalMsg.set_processcontrol(4);
	message::OutputDo outputDo;
	terminalMsg.mutable_outputdo()->CopyFrom(outputDo);
	std::string send_singlechip_data_str;
	google::protobuf::util::MessageToJsonString(terminalMsg ,&send_singlechip_data_str);

	//通知单片机开门
	m_singlechip_communication_vector[request.id_struct().terminal_id()].communication_tcp_send(send_singlechip_data_str.c_str(),send_singlechip_data_str.length());
	LOG(INFO) << "车牌号:" <<request.car_info().car_numberplate() << "      单元号:" << (char)('A'+request.id_struct().unit_id())  << "      出口:" <<(request.id_struct().terminal_id() + 1) << "      通知单片机开门----"  << this;
//	std::cout << " Msg = " << terminalMsg.DebugString() << std::endl;

//	LOG(INFO) << "车牌号:" <<request.car_info().car_numberplate() << "      单元号:" << (char)('A'+request.id_struct().unit_id())<< "      出口:" <<(request.id_struct().terminal_id() + 1) << "      LED屏提示取车已完成----"<< this;
	//检查门内是否有车 有车发送反馈
	message::Singlechip_data t_singlechip_data;
	{//上锁
		std::unique_lock<std::mutex> lk(m_singlechip_data_mutex);
		t_singlechip_data = m_singlechip_data_vector[request.id_struct().terminal_id()] ;
	}

	//如果门内有车发送反馈
	if(t_singlechip_data.insideexistenceflag() == 1)
	{
		//给主控发送反馈
		message::Notify_response_msg notifyResponseMsg;
		message::Base_info baseInfo;
		message::Error_manager errorManager;
		message::Id_struct idStruct;

		baseInfo.set_sender(message::Communicator::eNotify);
		baseInfo.set_receiver(message::Communicator::eMain);
		baseInfo.set_timeout_ms(1000);
		baseInfo.set_msg_type(message::Message_type::eNotify_response_msg);
		errorManager.set_error_code(0);
		idStruct.set_terminal_id(request.id_struct().terminal_id());

		notifyResponseMsg.mutable_base_info()->CopyFrom(baseInfo);
		notifyResponseMsg.mutable_code()->CopyFrom(errorManager);
		notifyResponseMsg.mutable_id_struct()->CopyFrom(idStruct);
		notifyResponseMsg.set_command_key(request.command_key());

		Notify_manage_communication::get_instance_references().encapsulate_msg(notifyResponseMsg.SerializeAsString());

		LOG(INFO) << "车牌号:" <<request.car_info().car_numberplate() << "      出口:" <<(request.id_struct().terminal_id() + 1) << "      车辆已送到出口，等待车辆离开----给主控发送反馈---"<< this;
	}
	else
	{
	    LOG(INFO) << " 消息错误！库内无车！ "<<t_singlechip_data.DebugString()<< this;
	}
}
//发送LED显示字符串
int Notify_manager::send_led_showstring(int area_index,std::string show_string)
{
	//将提示字符串转码设置
	m_led_protocol.set_show_string(area_index,String_convert::utf8_to_gbk(show_string));

	std::string t_tipstr = m_led_protocol.get_led_msg_total_string(area_index);
	//读取IP和端口。
	std::string t_ip = m_notify_communication_parameter_all.notify_communication_parameters().led_screen_ip();
	unsigned short t_port = m_notify_communication_parameter_all.notify_communication_parameters().led_screen_port();
	//初始化LED屏的通信
	Communication_tcp t_led_communication;						//LED屏通信
	t_led_communication.communication_tcp_init();
	int send_length = 0;
	std::chrono::system_clock::time_point	t_receive_time = std::chrono::system_clock::now();	//当前时间
	//连接LED屏幕
	while(	std::chrono::system_clock::now() - t_receive_time < (std::chrono::milliseconds)3000)
	{
		bool result = t_led_communication.communication_tcp_connect(t_ip.c_str(),t_port);
		if (  result )
		{
			send_length = t_led_communication.communication_tcp_send(t_tipstr.c_str(),t_tipstr.length());
			std::cout<< "--------  更新LED屏成功！ area_index = "<<area_index<<" 	show_string = "<<show_string<<" 	send_length = "<<send_length<<"  --------"<<std::endl;
			break;
//		LOG(INFO) << "--------  更新LED屏成功！ show_string = "<<show_string<<"  --------"<<this;
		}
		else
		{
			std::cout << "--------  LED屏连接失败! 更新失败！重试中---area_index = "<<area_index<<" 	show_string = "<<show_string<<"  --------"<<std::endl;
//		LOG(INFO) << "--------  LED屏连接失败! 更新失败！--------" <<this;
		}
	}
	//关闭连接
	t_led_communication.communication_tcp_uninit();

	return send_length;

}