/*
 * @Description: 车位分配算法模块，使用单例模式，接收外部请求并通过调用通信块接口发送反馈
 * @Author: yct
 * @Date: 2020-07-10 09:25:56
 * @LastEditTime: 2020-08-06 10:18:56
 * @LastEditors: yct
 */

#ifndef PARKSPACE_ALLOCATOR_HH
#define PARKSPACE_ALLOCATOR_HH

#include "../tool/thread_pool.h"
#include "../tool/singleton.h"
#include "../tool/thread_condition.h"
#include "../error_code/error_code.h"
#include "../communication/communication_message.h"
#include "../communication/communication.pb.h"
#include "../message/process_message.pb.h"
#include "../message/notify_message.pb.h"
#include "../communication/communication_tcp.h"
#include "../tool/proto_tool.h"
#include "../notify_manage/notify_communication.pb.h"
#include "../led/led_protocol.h"
#include "../led/led_outlet.h"
#include <google/protobuf/message.h>
#include <google/protobuf/util/json_util.h>
#include <google/protobuf/text_format.h>
#include "../message/dispatch_message.pb.h"

#include "string.h"
#include <list>
#include <vector>
#include <ctime>
#include <stdlib.h>
#include <mutex>
#include <iostream>
#include <fstream>
//Notify_manager
class Notify_manager : public Singleton<Notify_manager>
{
	friend class Singleton<Notify_manager>;
private:
	// 父类的构造函数必须保护，子类的构造函数必须私有。
	Notify_manager();
public:
	enum Notify_manager_satus
	{
		eNotify_manager_unknown,
		eNotify_manager_normal,
		eNotify_manager_fault
	};

#define SINGLE_CHIP_COMMUNICATION_PARAMETER_PATH "../setting/notify_manage_communication.prototxt"
#define UNIT_COUNT 3

#define  DATA_HEADER_LENGTH      4
public:
	// 必须关闭拷贝构造和赋值构造，只能通过 get_instance 函数来进行操作唯一的实例。
	Notify_manager(const Notify_manager &) = delete;
	Notify_manager &operator=(const Notify_manager &) = delete;
	~Notify_manager();

public://API functions
	//初始化
	Error_manager notify_manager_init();
	//从protobuf读取
	Error_manager notify_manager_init_from_protobuf(Communication_proto::Notify_communication_parameter_all& communication_parameter_all);
	//启动
	Error_manager notify_manager_run();
	//反初始化
	Error_manager notify_manager_uninit();
	//检查执行者的状态, 判断能否处理这条消息,
	Error_manager check_status();
	//处理消息的执行函数
	Error_manager execute_msg(Communication_message* p_msg);

	//数据处理
	std::string data_processing(char* data,int singlechip_id);
	//发送LED屏显示字符串 内部自动连接LED屏
	int send_led_showstring(int area_index,std::string show_string);

	//接受单片机数据线程
	void receive_singlechip_data_thread(int singlechip_id);
	//自动重连线程
	void reconnect_thread();

	//处理调度模块终端出入口的状态
	void execute_for_dispatch_terminal_status_msg(message::Dispatch_terminal_status_msg msg);
	//取车等候请求线程
	void execute_for_notify_request(message::Notify_request_msg request);

//	//等待车辆离开线程
//	void notify_car_departure_thread();



	void test();

public://execute_msg创建各线程进行处理
	//单片机有多个 所以对应资源也有多份
	std::vector<Communication_tcp>	 			m_singlechip_communication_vector; 			//单片机通信
	std::vector<std::thread*>					mp_receive_singlechip_data_thread_vector;	//接受的线程指针
	std::vector<Thread_condition*>				m_receive_condition_vector;					//接受的条件变量

	std::thread*								mp_reconnect_thread;						//重连线程
	Thread_condition							m_reconnect_condition;						//重连线程条件变量

	std::vector<message::Singlechip_data>		m_singlechip_data_vector;					//单片机数据
	std::mutex 									m_singlechip_data_mutex;					//单片机数据锁

	//调度模块终端出入口的状态
	std::vector<message::Dispatch_terminal_status_msg>		m_dispatch_terminal_status_msg_vector;		//调度模块终端出入口的状态
	std::vector<std::string>								m_dispatch_pickup_process_updata_time_vector;//调度流程更新时间

	Led_protocol								m_led_protocol;								//LED屏操作
	Notify_manager_satus 						m_notify_manager_satus;     				//取车等候区当前状态
	Thread_pool 								m_thread_pool;								//执行多任务的线程池

	Communication_proto::Notify_communication_parameter_all m_notify_communication_parameter_all;//配置文件所有参数  保存下来用于后续重连

};

#endif // !PARKSPACE_ALLOCATOR_HH

